/* Automatically generated nanopb constant definitions */
/* Generated by nanopb-0.4.8-dev */

#include "opentelemetry/resource.pb.h"
#if PB_PROTO_HEADER_VERSION != 40
#error Regenerate this file with the current version of nanopb generator.
#endif

PB_BIND(opentelemetry_proto_resource_v1_Resource, opentelemetry_proto_resource_v1_Resource, AUTO)



