/*
 * librdkafka - Apache Kafka C/C++ library
 *
 * Copyright (c) 2014-2022, Magnus Edenhill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <string>

#include "rdkafkacpp_int.h"

int RdKafka::version() {
  return rd_kafka_version();
}

std::string RdKafka::version_str() {
  return std::string(rd_kafka_version_str());
}

std::string RdKafka::get_debug_contexts() {
  return std::string(RD_KAFKA_DEBUG_CONTEXTS);
}

std::string RdKafka::err2str(RdKafka::ErrorCode err) {
  return std::string(rd_kafka_err2str(static_cast<rd_kafka_resp_err_t>(err)));
}

int RdKafka::wait_destroyed(int timeout_ms) {
  return rd_kafka_wait_destroyed(timeout_ms);
}

void *RdKafka::mem_malloc(size_t size) {
  return rd_kafka_mem_malloc(NULL, size);
}

void RdKafka::mem_free(void *ptr) {
  rd_kafka_mem_free(NULL, ptr);
}
